__all__ = ['start_wsgi_server']

from _wsgi import handler as wsgi_app

try:
    from _profile import wsgi_profile
    wsgi_app = wsgi_profile(wsgi_app)
    print "wsgi profiler path: /__profiler__"
except ImportError:
    pass

try:
    from cherrypy import wsgiserver as cpws

    def start_wsgi_server(host = '127.0.0.1', port = 9000, app = wsgi_app):
        print "serving on http://%s:%s" % (host, port)
        s = cpws.CherryPyWSGIServer((host, port), app, request_queue_size = 500)
        try:
            s.start()
        except KeyboardInterrupt:
            s.stop()

except ImportError:
    try:
        from paste import httpserver

        def start_wsgi_server(host = '127.0.0.1', port = 9000, app = wsgi_app):
            httpserver.serve(app, host, port, request_queue_size = 500)

    except ImportError:
        from wsgiref.simple_server import make_server

        def start_wsgi_server(host = '127.0.0.1', port = 9000, app = wsgi_app):
            make_server(host, port, app).serve_forever()

#NOTE: for production use (gunicorn|uwsgi)+nginx+supervisor



from time import time as timestamp
from os.path import splitext
from webob import Request
import requests

from _term import parse_term, parse_json, to_simple, to_json
from _crypto import AES, hmac, gen_token
from _auth import State
from _service import consume
from _wsgi import handler as wsgi_app

class PyClient(object):

    def __init__(self):
        self.__state = State()

    def _consume(self, urlpath, **argd):
        _, ext, ret = consume(self.__state, urlpath, **argd)
        return ext, ret

    def call(self, urlpath, **argd):
        ext, ret = self._consume(urlpath, **argd)
        return parse_json(ret) if ext == '.json' else ret

    def safe_call(self, urlpath, **argd):
        state = self.__state
        sid, token = str(state.sid), state.token
        params = to_json(to_simple(argd))
        sign = hmac(token, params)
        aes = AES(token)
        enc = aes.encrypt('|'.join((sign, params)))
        enc_ret = self.call(urlpath, enc = '|'.join((sid, enc)))
        ret = aes.decrypt(enc_ret)
        sign, new_token, json = ret.split('|', 2)
        assert sign == hmac(token, '|'.join((new_token, sid, json))), \
                ('bad-signature', sid)
        state.update(token = new_token)
        return parse_term(parse_json(json))

    def login(self, login, password, force = False, **extra):
        nonce = gen_token()[:8]
        token = gen_token()
        data = '%s|%s|%s' % (int(timestamp()), nonce, token)
        enc = AES(nonce + password).encrypt(data)
        enc_ret = self.call('login.json',
                uid = login, nonce = nonce, enc = enc,
                force = force, **extra)
        ret = AES(token).decrypt(enc_ret)
        self.__state.update(parse_term(parse_json(ret)))
        return True

    def logout(self):
        return self.safe_call('logout.json')

class WsgiClient(PyClient):

    def _consume(self, urlpath, **argd):
        _, ext = splitext(urlpath)
        if argd:
            params = dict((k, (v.name, v) if isinstance(v, file) else v)
                          for k, v in argd.iteritems())
            req = Request.blank(urlpath, POST = params)
        else:
            req = Request.blank(urlpath)
        #req.charset = 'utf8'
        res = req.get_response(wsgi_app)
        return ext, res.body

class HttpClient(PyClient):

    def __init__(self, baseurl = 'http://127.0.0.1:9000'):
        super(HttpClient, self).__init__()
        self.__baseurl = baseurl

    def _consume(self, urlpath, **argd):
        url = '/'.join((self.__baseurl, urlpath))
        _, ext = splitext(url)
        if argd:
            data, files = dict(), dict()
            for k, v in argd.iteritems():
                if isinstance(v, file):
                    files[k] = v
                else:
                    data[k] = v
            res = requests.post(url, files=files, data=data)
        else:
            res = requests.get(url)
        return ext, res.text

if __name__ == '__main__':
    client = PyClient()
    client.login('admin', 'admin')



"""
Fanery is a Python framework that offers:

    - minimalist pythonic design.
    - strong security.
    - easy of use.
    - scalable fundation.
    - multi-protocol facility.
    - handy profiling support.

Fanery philosofy:

    - Secure by default (yes I love OpenBSD).
    - Free/Open Source Software is the Way.
    - UI should never be created server-side.
    - functional style programming should be
      the preferred way to build applications.

Required dependencies:

    - pycrypto.
    - webob.

Optional dependencies:

    - memory-profiler.
    - line-profiler.
    - profilehooks.
    - pprofiler.
    - linesman.
    - rfoo.
    - lmdb.

Getting started:

    Creating secure services:

        from fanery import service

        @service()
        def safe_call(*args, **argd):
            '''safe_call service.'''
            return dict(args = args, argd = argd)

    Exposed services are pure Python functions,
    don't have to care about client nature:

        - Is it a browser? a raw socket consumer? etc.

    By default call parameters (*args and **argd) are parsed safely
    to python objects by fanery.parse_term(). This behaviour can be
    disabled by setting auto_parse to False:

        @service(auto_parse = False)
        def safe_call(...)

    Exposing hello service through WSGI:

        from fanery import wsgi_server

        # bind by default to localhost:9000
        wsgi_server().start()

Fanery @service decorator explained:

    The @service decorator wraps a function in such
    a way that by default can be consumed only if:

        - The caller own a valid session (is autenticated).
        - The remote call is made through SSL/TLS secured channel.
        - The remote call is encrypted as follow:
            - provide unique session ID and valid HMAC call signature.
            - encrypt the call with OpenSSL compatible AES encryption
              using the server generated cryptographic session token
              which is unique for each single call.

    Such strong security requirements are enforced by the @service
    decorator to avoid and protect applications from attacks like:

        - Cross Site Request Forgery (CSRF).
        - Session Hijacking.
        - Session poisoning.
        - Unwanted redirection.
        - Eyesdrop even in the presence of SSL/TLS weakness.
        - Intercept/resend call stream.


    Aditionally for the sake of remote call safety and application
    scalability fanery:

        - Default login() implementation never send user passwords.
        - Do not rely on session ID secrecy for security, in other
          words, security will not be compromised if session ID is
          disclosed.
        - Bind session to remote call execution stack; session data
          does not rely on thread local facility but fanery.get_state()
          behave like it was.

    When services are exposed via WSGI fanery also take care of
    enforcing safety by:

        - Never use cookies or browser local storage.
        - Never rely on ugly/obfuscated url.

Fanery @service decorator by example:

    from fanery import service, is_number

    @service('sum'):
    def sum_values(*num):
        '''
        sum_values function is exposed as "sum".

        Clients can call it in many ways:

            - with any defined output format:

                sum.json    (default)
                sum.txt     (use %s)
                sum.raw     (use %r)

            - with any prefix:
                prefix/sum.json
                multy/path/prefix/sum.txt

            - with arguments in urlpath:

                sum/n1/n2/n3.raw

            - with GET/POST request arguments:

                GET sum.json?num=1&num=2&num=-4

                POST prefix/sum.txt
                num=1&num=2&num=-4
        '''
        return sum(n for n in num if is_number(n))
"""




User = 'User'

class MemDictStore(object):

    transactional = False

    def __init__(self, auth_hooks = None):
        self._models = dict()
        self._sessions = dict()
        if isinstance(auth_hooks, dict):
            auth_hooks.update(load_state = self.load_state,
                              store_state = self.store_state,
                              destroy_state = self.destroy_state,
                              user_sessions = self.user_sessions,
                              fetch_user = self.fetch_user,
                              get_user = self.get_user)

    def add_model(self, model, indexer):
        self._models[model] = dict()

    def upsert(self, txn, record):
        self._models[record._model][record._uuid] = record
    insert = update = upsert

    def delete(self, txn, record):
        del self._models[record._model][record._uuid]

    def fetch(self, model, uuid, vsn = None, txn = None):
        record = self._models[model][uuid]
        assert not vsn or record._vsn == vsn
        assert not txn or record._txn == txn
        return record._vsn, record._txn, record._dct

    def query(self, model, fun, score = None):
        for record in filter(fun, self._models[model].itervalues()):
            uuid, vsn, txn = record._uuid, record._vsn, record._txn
            yield score is not None and (uuid, vsn, txn, score) \
                                     or (uuid, vsn, txn)

    def search(self, model, fun):
        return self.query(model, fun, 1)

    def select(self, model, fun):
        return self.query(model, fun)

    def get_user(self, uuid):
        return self._models[User][uuid]

    def user_sessions(self, uuid):
        return len(self._sessions.get(uuid, None) or {})

    def fetch_user(self, uid, **extra):
        for record in self._models[User].itervalues():
            if record.login == uid:
                if not extra:
                    return record
                else:
                    dct = record._dct
                    for k, v in extra.iteritems():
                        if k in dct and dct[k] == v:
                            continue
                        else:
                            break
                    else:
                        return record

    def load_state(self, sid):
        return self._sessions.get(sid, None) or {}

    def store_state(self, state):
        self._sessions[state.sid] = state

    def destroy_state(self, state):
        del self._sessions[state.sid]




__all__ = ['service', 'static']

from _term import parse_term, parse_json, to_simple, to_json, to_str, is_str
from _auth import load_state, store_state, get_state, hooks
from _crypto import AES, hmac, gen_token

from os.path import splitext
from functools import wraps
import logging
import cgitb
import sys

registry = dict()

def decrypt(enc):
    assert enc and is_str(enc) and enc.count('|') == 1, ('bad-format', enc)
    sid, enc = enc.split('|', 1)
    sid = parse_term(sid)
    _state_ = load_state(sid)
    assert _state_.token, ('null-token', sid, _state_)
    aes = AES(_state_.token)
    sign, json = aes.decrypt(enc).split('|', 1)
    assert sign == hmac(_state_.token, json), ('bad-signature', sid)
    return _state_, parse_json(json)

def encrypt(_state_, term):
    assert _state_.token, ('null-token', _state_.sid)
    token, json = _state_.token, to_json(to_simple(term))
    new_token = gen_token() if hooks.ONE_TIME_TOKEN else token
    sign = hmac(token, '|'.join((new_token, str(_state_.sid), json)))
    aes = AES(token)
    enc = aes.encrypt('|'.join((sign, new_token, json)))
    _state_.token = new_token
    if _state_.expires > 0:
        store_state(_state_)
    return enc

def service(urlpath = None, auto_parse = True, static = False, cache = False, \
            force_download = False, safe = True, ssl = True, log_exc = True, \
            _REMOVE_force = False, **filters):

    assert urlpath not in registry or force, ('already-defined', urlpath)

    filters.setdefault('json', to_json)
    filters.setdefault('txt', to_str)
    filters.setdefault('raw', repr)

    def decorator(f):
        key = (urlpath or f.__name__).strip('/')
        for k, v in filters.iteritems():
            assert k and k.isalnum() and callable(v), ('bad-filter', k, v)

        @wraps(f)
        def wrapper(*args, **argd):
            _state_ = get_state()
            if ssl is True:
                assert _state_.ssl, ('ssl-required', urlpath)
            if safe is True:
                enc = args[0] if args else argd.pop('enc', None)
                state, params = decrypt(enc)
                _state_.update(state)
                argd.update(params)
            if auto_parse is True:
                args = parse_term(args) if args else args
                argd = parse_term(argd) if argd else argd
            try:
                ret = f(*args, **argd)
            except Exception, e:
                if log_exc is True:
                    logging.error(cgitb.text(sys.exc_info()))
                if safe is True:
                    return encrypt(_state_, dict(exc = e.args))
                else:
                    raise
            else:
                return encrypt(_state_, ret) if safe is True else ret

        wrapper.ssl = ssl
        wrapper.safe = safe
        wrapper.urlpath = key
        wrapper.cache = cache
        wrapper.static = static
        wrapper.log_exc = log_exc
        wrapper.filters = filters
        wrapper.auto_parse = auto_parse
        wrapper.force_download = force_download

        return registry.setdefault(key, wrapper)

    return decorator

def static(urlpath, root, index = None, **argd):

    from os.path import normpath, realpath, isdir, isfile, join

    root = join(realpath(normpath(root)), '')
    assert isdir(root), ('no-such-directory', root)

    @service(urlpath, auto_parse = False, static = True, safe = False, **argd)
    def serve_file(*args, **argd):
        filepath = realpath(normpath(join(root, *args)))
        if isdir(filepath) and index:
            filepath = join(filepath, index)
        assert isfile(filepath) and filepath.startswith(root), \
                ('not-found', filepath)
        return filepath

    return serve_file

acl_rules = dict(role = lambda s, roles, *args, **argd: s.role in roles,
                 user = lambda s, users, *args, **argd: s.login in users)

def set_acl_rule(name, func):
    acl_rules[name] = func

def acl(**rules):

    for k, v in rules.iteritems():
        validator = acl_rules.get(k, v)
        assert callable(validator), ('invalid-acl', k)
        rules[k] = (v, validator)

    def decorator(f):
        assert hasattr(f, 'urlpath'), ('not-a-service', f)

        @wraps(f)
        def wrapper(*args, **argd):
            _state_ = get_state()
            for k, (v, valid) in rules.iteritems():
                assert valid(_state_, v, *args, **argd), ('unauthorized', k, v)
            return f(*args, **argd)

        return wrapper

    return decorator

def lookup(urlpath):

    path, ext = splitext(urlpath.strip().strip('/'))
    args = []

    while path not in registry:
        try:
            path, arg = path.rsplit('/', 1)
            args.append(arg)
        except ValueError:
            break

    fun = registry.get(path, None)

    if not fun:
        try:
            prefix, urlpath = urlpath.split('/', 1)
            return lookup(urlpath)
        except ValueError:
            pass

    out = fun.filters.get(ext[1:], False) if ext and fun else None

    if args:
        args.reverse()
        if fun and fun.static is True:
            args[-1] += ext

    return fun, args, ext, out

def consume(_state_, urlpath, *args, **argd):
    fun, argv, ext, out = lookup(urlpath)

    assert callable(fun), ('not-found', urlpath)
    if ext and not fun.static:
        assert callable(out), ('bad-format', ext)

    argv.extend(args)
    ret = fun(*argv, **argd)
    return fun, ext, out(ret) if ext else ret



__all__ = ['login', 'logout', 'get_state', 'User']

from time import time as timestamp
from inspect import currentframe

from _term import Hict, gen_uuid, to_json, to_simple
from _crypto import AES, gen_token

class State(Hict): pass

def get_state():
    frame = current = currentframe()
    while frame:
        _state_ = frame.f_locals.get('_state_', None)
        if isinstance(_state_, State):
            if current is not frame:
                current.f_locals['_state_'] = _state_
            return _state_
        else:
            frame = frame.f_back
    return current.f_locals.setdefault('_state_', State())

def load_state(sid, **data):
    _state_ = State(hooks.load_state(sid))
    if _state_:
        assert _state_.expires > timestamp(), 'session-timeout: %s' % sid
        _state_.user = hooks.get_user(_state_.user)
    else:
        default = dict(origin = None, domain = 'localhost', sid = sid,
                       user = None, role = None, data = State(),
                       expires = timestamp() + hooks.SESSION_TIMEOUT,
                       token = None, timeout = hooks.SESSION_TIMEOUT)
        default.update(data)
        _state_.update(default)
        store_state(_state_)
    return _state_

def store_state(state):
    state.expires = timestamp() + hooks.SESSION_TIMEOUT
    state.user = getattr(state.user, '_uuid', None)
    hooks.store_state(state)

def _undefined(msg):
    def undefined(*args, **argd):
        raise AssertionError(msg)
    return undefined

hooks = Hict(load_state = _undefined('undefined auth.hooks.load_state'),
             store_state = _undefined('undefined auth.hooks.store_state'),
             destroy_state = _undefined('undefined auth.hooks.destroy_state'),
             user_sessions = _undefined('undefined auth.hooks.user_sessions'),
             fetch_user = _undefined('undefined auth.hooks.fetch_user'),
             get_user = _undefined('undefined auth.hooks.get_user'),
             MAX_ACTIVE_USER_SESSIONS = 1,
             SESSION_TIMEOUT = 900,
             ONE_TIME_TOKEN = True)

from _service import service

@service(safe = False, ssl = False, auto_parse = False, log_exc = False)
def login(uid, nonce, enc, force = False, **extra):
    user = hooks.fetch_user(uid, **extra)
    assert user, ('invalid-uid', uid)
    data = AES(nonce + user.password).decrypt(enc)
    assert data.count('|') == 2, ('bad-format', data)
    tstamp, _nonce, client_token = data.split('|', 2)
    assert tstamp.isdigit() and (timestamp() - int(tstamp)) < 1800, \
            ('invalid-timestamp', tstamp)
    assert _nonce == nonce, ('invalid-password', uid)
    if force is not True:
        active_user_sessions = hooks.user_sessions(user._uuid)
        assert active_user_sessions < hooks.MAX_ACTIVE_USER_SESSIONS, \
                ('active-sessions', active_user_sessions, user._uuid, sessions)
    sid, token = gen_uuid(), gen_token()
    _state_ = load_state(sid)
    assert not _state_.token, 'unexpected-state'
    _state_.update(user = user, role = user.role or user.login, token = token)
    _state_.data.update(extra)
    store_state(_state_)
    answer = dict(sid = sid, token = token, role = _state_.role)
    return AES(client_token).encrypt(to_json(to_simple(answer)))

@service(ssl = False, auto_parse = False, log_exc = True)
def logout(*args, **argd):
    _state_ = get_state()
    _state_.expires = 0
    hooks.destroy_state(_state_)
    return True



from inspect import ismodule as is_module, \
                    isfunction as is_function, \
                    isgenerator as is_generator, \
                    isbuiltin as is_builtin, \
                    isclass as is_class

from datetime import datetime, date
from decimal import Decimal
from uuid import UUID

class Hict(dict):
    """
    Hierarchical dotted dictionary.
    """

    def __missing__(self, key):
        term = self[key] = Hict()
        return term

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

def is_str(term):
    return isinstance(term, basestring)

def is_number(term):
    return isinstance(term, (int, float, Decimal))

def is_date(term):
    return isinstance(term, (date, datetime))

def is_uuid(term):
    return isinstance(term, UUID)

def is_sequence(term):
    return hasattr(term, '__iter__') and \
            not isinstance(term, (basestring, dict))

def is_dict(term):
    return isinstance(term, dict) or type(term) is dict

def is_inet_address(term):
    raise NotImplementedError

def is_inet6_address(term):
    raise NotImplementedError

def is_file(term):
    raise NotImplementedError

def is_dir(term):
    raise NotImplementedError

from re import compile as regex

from uuid import uuid4 as gen_uuid

try:
    from validate_email import validate_email as is_email
except ImportError:
    from email.utils import parseaddr as _parse_email_addr
    # borrowed from http://www.regular-expressions.info/email.html
    _email_regex = regex(r'''[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?''')
    def is_email(term, verify = False):
        try:
            name, email = _parse_email_addr(term)
            return _email_regex.match(email)
        except:
            return False

def to_str(term):
    #FIXME: handle any term
    return '%s' % term

def to_simple(term):
    if not term or isinstance(term, (int, float, basestring)):
        return term
    elif isinstance(term, dict) or type(term) is dict:
        return dict((to_simple(k), to_simple(v))
                    for k, v in term.iteritems())
    elif hasattr(term, '__iter__'):
        return [to_simple(t) for t in term]
    else:
        return str(term)

def parse_term(term, dct=dict):
    if not term:
        return term
    elif isinstance(term, basestring):
        try:
            f = float(term)
            i = int(f)
            if i == f:
                return i
            elif str(f) == term:
                return f
            else:
                return Decimal(term)
        except:
            pass
        try:
            return UUID(term)
        except:
            pass
        if len(term.strip()) >= 3:
            try:
                return parse_date(term)
            except:
                pass
        return term
    elif isinstance(term, dict) or type(term) is dict:
        return dct((parse_term(k, dct), parse_term(v, dct))
                    for k, v in term.iteritems())
    elif hasattr(term, '__iter__'):
        return type(term)(parse_term(t, dct) for t in term)
    else:
        return term

try:
    from dateutil.parser import parse as parse_date
except:
    raise #FIXME: multiple modules or simple default implementation

try:
    from ujson import dumps as to_json, \
                      loads as parse_json
except ImportError:
    try:
        from yajl import Encoder, Decoder
        parse_json = Decoder().decode
        to_json = Encoder().encode
    except ImportError:
        try:
            from jsonlib import write as to_json, \
                                read as parse_json
        except ImportError:
            try:
                from cjson import encode as to_json, \
                                  decode as parse_json
            except ImportError:
                try:
                    from simplejson import dumps as to_json, \
                                           loads as parse_json
                except ImportError:
                    from json import dumps as to_json, \
                                     loads as parse_json

from unicodedata import normalize as _normalize
_punct_regex = regex(r'[\t !"#$%&\'()*\-/<=>?@\[\\\]^_`{|},.:]+')

def normalize(term, mode = 'NFKD'):
    assert isinstance(term, basestring), 'bad-type: %r' % term
    text = ' '.join(_punct_regex.split(text))
    return _normalize(mode, text).encode('ascii', 'ignore')

def slugify(term, delim = '-'):
    return normalize(term).lower().replace(' ', delim)



"""
Handle sessions in execution stack.
"""
__all__ = ['get_state']

from time import time as timestamp
from inspect import currentframe

from _term import Hict

class State(Hict): pass

def get_state():
    frame = current = currentframe()
    while frame:
        _state_ = frame.f_locals.get('_state_', None)
        if isinstance(_state_, State):
            if current is not frame:
                current.f_locals['_state_'] = _state_
            return _state_
        else:
            frame = frame.f_back
    return current.f_locals.setdefault('_state_', State())

def load_state(sid, **data):
    _state_ = hooks.load_state(sid)
    if _state_:
        assert _state_.expires > timestamp(), 'session-timeout: %s' % sid
        _state_.user = auth_hooks.get_user(_state_.user)
    else:
        default = dict(origin = None, domain = 'localhost', sid = sid,
                       user = None, role = None, data = State(),
                       expires = timestamp() + hooks.SESSION_TIMEOUT,
                       token = None, timeout = hooks.SESSION_TIMEOUT)
        default.update(data)
        _state_ = State(default)
        store_state(_state_)
    return _state_

def store_state(state):
    state.expires = timestamp() + hooks.SESSION_TIMEOUT
    state = State(state)
    state.user = getattr(state.user, '_uuid', None)
    hooks.store_state(state)

def clear_state(state):
    state.data.clear()
    state.expires = 0

def user_sessions(uid):
    return hooks.user_sessions(uid)

sessions = dict()

hooks = Hict(load_state = lambda sid: sessions.get(sid, None),
             store_state = lambda state: sessions.__setitem__(state.sid, state)\
                        if state.expires else sessions.__delitem__(state.sid),
             user_sessions = lambda uid: sum(v.user == uid for v in
                                             sessions.itervalues()),
             SESSION_TIMEOUT = 900)

from _auth import hooks as auth_hooks



"""
Storage proxy, not an ORM.
"""
__all__ = ['add_storage', 'add_model', 'Aict', 'Record', 'storage']

from _term import to_simple, parse_term, gen_uuid

storage_reg = dict()
model_reg = dict()

def add_storage(storage, *models):
    for model in models:
        assert model in model_reg, ('unknown-model', model)
        storage_reg[model] = storage
        storage.add_model(model, get_indexer(model))

def add_model(model, validator = None, initializer = None, indexer = None):
    model_reg[model] = (validator, initializer, indexer)

def get_storage(model):
    assert model in storage_reg, model
    return storage_reg[model]

def get_validator(model):
    return model_reg[model][0] if model in model_reg else None

def get_initializer(model):
    return model_reg[model][1] if model in model_reg else None

def get_indexer(model):
    return model_reg[model][2] if model in model_reg else None

class Aict(dict):
    """
    Hierarchical dotted dictionary with value auto-parsing.
    """

    def __missing__(self, key):
        term = self[key] = Aict()
        return term

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, parse_term(value, dct=Aict))

    __setattr__ = __setitem__
    __getattr__ = dict.__getitem__
    __delattr__ = dict.__delitem__

class Record(object):

    def __init__(self, model, uuid, vsn, txn = None, _dct_ = None, **dct):
        store = get_storage(model)
        if _dct_:
            self._dct = parse_term(_dct_, dct=Aict)
        elif dct:
            self._dct = parse_term(dct, dct=Aict)
        else:
            record_data = store.get(uuid, txn, vsn, txn)
            assert record_data, 'not-found: %s %s' % (model, uuid)
            uuid, vsn, txn, data = record_data
            self._dct = parse_term(data, dct=Aict)
        self._mark = 0 # 1 -> insert, 2 -> update, 3 -> delete
        self._store = store
        self._model = model
        self._uuid = parse_term(uuid)
        self._vsn = parse_term(vsn)
        self._txn = parse_term(txn)
        initialize = get_initializer(model)
        if initialize:
            initialize(self)

    def __getattr__(self, key):
        return self._dct[key]

    def __setattr__(self, key, value):
        dct = self.__dict__ if key[0] == '_' else self._dct
        dct[key] = parse_term(value, dct=Aict)

    def __delattr__(self, key):
        if key[0] != '_':
            del self._dct[key]

    def update(self, *args, **argd):
        self._dct.update(*args if args else argd)
        return self

    def merge(self, *args, **argd):
        dct = self._dct
        dct.update((k, v) for k, v in (args or argd.iteritems()) if k not in dct)
        return self

class storage(object):

    def __init__(self):
        self._cache = dict()
        self._txn = gen_uuid()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self._cache:
            if isinstance(value, Exception):
                self.__rollback()
            else:
                self.__commit()

    def __commit(self):
        dirty, dbms = set(), set()
        for (model, uuid, vsn), record in self._cache.iteritems():
            if record._mark:
                store = record._store
                if store.transactional is True:
                    dbms.add(store)
                dirty.add(record)
        try:
            txn = self._txn
            for dbm in dbms:
                dbm.begin_transaction(txn)
            for record in dirty:
                mark = record._mark
                try:
                    _vsn, _txn = record._vsn, record._txn
                    if mark == 1:
                        record._vsn = 1
                        record._store.insert(txn, record)
                    elif mark == 2:
                        record._vsn += 1
                        record._store.update(txn, record)
                    elif mark == 3:
                        record._vsn *= -1
                        record._store.delete(txn, record)
                    record._txn = txn
                    record._mark = 0
                except:
                    record._vsn, record._txn = _vsn, _txn
                    raise
            for dbm in dbms:
                dbm.commit(txn)
        except:
            self.__rollback()

    def __rollback(self):
        txn, dbms = self._txn, set()
        for record in self._cache.itervalues():
            if record._mark:
                store = record._store
                if store.transactional is True:
                    dbms.add(store)
        for dbm in dbms:
            dbm.rollback(txn)

    def fetch(self, model, uuid, vsn = None, txn = None, acl = None, fail = None):
        cache = self._cache
        record = cache.get((model, uuid, vsn), None)
        if record and record._mark == 3:
            record = None
        elif record is None:
            storage = get_storage(model)
            term = storage.fetch(model, uuid, vsn, txn)
            record = term if isintance(Record, term) else Record(model, uuid, *term)
            cache[(model, uuid, record._vsn)] = record

        if record and acl and acl(record) is not True:
            record = None

        if record:
            return record
        elif issubclass(fail, Exception):
            raise fail
        elif callable(fail):
            return fail()
        else:
            return fail

    def query(self, method, model, fun, acl = None, raw = False, load = parse_term):
        index = get_storage(model)
        cache = self._cache

        for term in getattr(index, method)(model, fun):
            if raw is True:
                yield load(term) if load else term

            if len(term) == 4:
                uuid, vsn, txn, score = term
            else:
                uuid, vsn, txn = term
                score = None
            record = cache.get((model, uuid, vsn), None)

            if record is None:
                record = self.fetch(model, uuid, vsn, txn, acl)
                if record is None:
                    continue
                cache[(model, uuid, vsn)] = record
            elif record._mark == 3 or (acl and acl(record) is not True):
                continue
            yield (score, record) if score is not None else record

    def search(self, model, fun, acl = None):
        return self.query('search', model, fun, acl)

    def select(self, model, fun, acl = None):
        return self.query('select', model, fun, acl)

    def select_one(self, model, fun, acl = None):
        record = None
        for idx, record in enumerate(self.select(model, fun, acl)):
            assert idx == 0, 'multiple-results'
        return record

    def insert(self, model, *args, **argd):
        uuid, vsn = gen_uuid(), 0
        record = Record(model, uuid, vsn, self._txn, *args, **argd)
        self.validate(record)
        record._mark = 1
        return self._cache.setdefault((model, uuid, vsn), record)

    def update(self, *records):
        for record in records:
            mark = record._mark
            assert mark != 3 and record._vsn >= 0, record
            self.validate(record)
            if mark == 0:
                record._mark = 2

    def delete(self, *records):
        for record in records:
            assert record._mark != 1 and record._vsn > 0, record
            record._mark = 3

    def validate(self, record):
        validator = get_validator(record._model)
        if validator:
            errors = validator(record)
            assert not errors, (record, errors)



"""
Usefull profiling decorators and WSGI middleware.
"""
__all__ = ['timecall', 'wsgi_profile', 'memory_profile', 'line_profile', 'coverage']

from functools import wraps
from logging import info

try:
    from linesman.middleware import make_linesman_middleware as wsgi_profile
except ImportError:
    pass

try:
    from memory_profiler import profile as memory_profile
except ImportError:
    pass

try:
    from line_profiler import profile as line_profile
except ImportError:
    try:
        from pprofile import Profile

        def line_profile(f, match = None):
            @wraps(f)
            def wrapper(*args, **argd):
                try:
                    profiler = Profile()
                    with profiler:
                        return f(*args, **argd)
                finally:
                    fdict = profiler.file_dict
                    for path in fdict.iterkeys():
                        if 'python' in path:
                            del fdict[path]
                    profiler.print_stats()
            return wrapper
    except ImportError:
        pass

try:
    from profilehooks import profile
except ImportError:
    try:
        from pstats import Stats

        try:
            from cProfile import Profile
        except ImportError:
            from profile import Profile

        def profile(f, immediate = True):
            @wraps(f)
            def wrapper(*args, **argd):
                profiler = Profile()
                try:
                    profiler.enable()
                    return f(*args, **argd)
                finally:
                    profiler.disable()
                    stats = Stats(profiler)
                    stats.strip_dirs()
                    stats.sort_stats('cumulative')
                    stats.print_stats(f.__name__)
            return wrapper
    except ImportError:
        pass

try:
    from profilehooks import timecall
except ImportError:
    from time import time
    def timecall(f, immediate = True):
        @wraps(f)
        def wrapper(*args, **argd):
            try:
                start = time()
                return f(*args, **argd)
            finally:
                print "Execution time: %s s (%s)" % (time() - start, f)
        return wrapper

try:
    from profilehooks import coverage
except ImportError:
    pass



from mimetypes import guess_type, add_type
add_type('application/json', '.json')

from webob import Response
from webob.dec import wsgify
from webob.exc import HTTPServerError, HTTPBadRequest
from webob.static import FileApp, DirectoryApp

from _term import (
    Hict, parse_json, to_simple, to_json,
    is_file, is_dir, is_generator
)
from _service import consume
from _auth import State

CONTENT_TYPE = 'text/plain'
SERVER = 'FWS/1.0'
CHARSET = 'utf8'

@wsgify
def handler(req):
    _state_ = State(domain = req.host.split(':', 1)[0],
                    ssl = req.scheme == 'https',
                    origin = req.remote_addr,
                    user = None, role = None,
                    sid = None, token = None,
                    data = State())

    req.charset = CHARSET

    argd, params = Hict(), req.params
    for key in params.iterkeys():
        if key.endswith('[]'):
            argd[key[:-2]] = params.getall(key)
        else:
            argd[key] = params[key]

    if req.is_xhr:
        argd.update(parse_json(req.json_body))

    try:
        fun, ext, ret = consume(_state_, req.path, **argd)
    except Exception, e:
        error = to_json(dict(error = to_simple(*e.args)))
        try:
            raise
        except AssertionError:
            return HTTPBadRequest(error)
        except:
            return HTTPServerError()

    if fun.static:
        if is_file(ret):
            res = FileApp(ret)
        elif is_dir(ret):
            res = DirectoryApp(ret)
    elif isinstance(ret, file):
        with ret:
            res = FileApp(ret.name)
    else:
        content_type = guess_type(fun.urlpath+ext)[0] or CONTENT_TYPE
        res = Response(charset = CHARSET, content_type = content_type)

        if isinstance(ret, str):
            res.body = ret
        elif isinstance(ret, unicode):
            res.unicode_body = ret
        elif isinstance(ret, file):
            res.body_file = ret
        elif isinstance(ret, (tuple, list)) or is_generator(ret):
            res.app_iter = ret
        else:
            res.body = repr(ret)

    if not fun.cache:
        res.cache_expires(0)

    res.server = SERVER

    return res



from fanery import service, static, is_date

@service(safe = False, ssl = False)
def echo(*args, **argd):
    if args and argd:
        return args, argd
    elif args:
        return args if len(args) > 1 else args[0]
    elif argd:
        return argd

@service('echo_two', safe = False, ssl = False, auto_parse = False)
def echo2(*args, **argd):
    if args and argd:
        return args, argd
    elif args:
        return args if len(args) > 1 else args[0]
    elif argd:
        return argd

@service(ssl = False)
def safe_call(**argd):
    return argd

def test_urlpath():
    assert echo.urlpath == 'echo'
    assert echo2.urlpath == 'echo_two'
    assert safe_call.urlpath == 'safe_call'

def test_safe_param():
    assert echo.safe is False
    assert echo2.safe is False
    assert safe_call.safe is True

def test_ssl_param():
    assert echo.ssl is False
    assert echo2.ssl is False
    assert safe_call.ssl is False

def test_auto_parse_param():
    assert echo.auto_parse is True
    assert echo2.auto_parse is False
    assert safe_call.auto_parse is True

def test_log_exc_param():
    assert echo.log_exc is True
    assert echo2.log_exc is True
    assert safe_call.log_exc is True

def test_auto_parse_on_call():
    assert echo('1') == 1
    assert echo('1.1') == 1.1
    assert is_date(echo('1990-01-01'))
    assert echo2('1') == '1'
    try:
        safe_call()
    except AssertionError, e:
        assert e.args[0][0] == 'bad-format', e.args

def test_lookup():
    from fanery._term import to_json, to_str
    from fanery._service import lookup

    assert lookup('prefix/echo/1/2/3.json') == (
            echo, ['1', '2', '3'], '.json', to_json)

    assert lookup('/echo.doc') == (echo, [], '.doc', False)
    assert lookup('/a/safe_call') == (safe_call, [], '', None)
    assert lookup('a/b/echo_two.raw') == (echo2, [], '.raw', repr)
    assert lookup('safe_call.txt') == (safe_call, [], '.txt', to_str)

def test_consume():
    from fanery._state import get_state
    from fanery._service import consume

    _state_ = get_state()
    assert consume(_state_, 'prefix/echo/1/2/3') == (echo, '', (1,2,3))
    assert consume(_state_, 'prefix/echo/1/2/3.json') == (echo, '.json', '[1, 2, 3]')

def store_setUp():
    from fanery._store import storage, add_model, add_storage
    from fanery._mddb import MemDictStore as Store, User
    from fanery._auth import hooks

    add_model(User)
    dbm = Store(hooks)
    add_storage(dbm, User)

    user = hooks.fetch_user('user')
    if not user:
        with storage() as db:
            user = db.insert(User, login = 'user', password = 'passwd')
            assert user.login == 'user'
            assert user.password == 'passwd'

        stored = hooks.fetch_user('user')
        assert user.__dict__ == stored.__dict__

    return User, user

def test_PyClient():
    from fanery._client import PyClient

    client = PyClient()
    User, user = store_setUp()

    assert client.login(user.login, user.password) is True
    assert client.call('echo', a = '1.0') == {'a': 1}
    assert client.safe_call('safe_call', a = 1) == {'a': 1}
    assert client.call('echo.json', a = '1.0') == {'a': 1}
    assert client.safe_call('safe_call.json', a = '1.0') == {'a': 1}
    assert client.logout() is True

def test_WsgiClient():
    from fanery._client import WsgiClient

    client = WsgiClient()
    User, user = store_setUp()

    assert client.login(user.login, user.password) is True
    assert client.call('echo', a = '1.0') == repr({'a': 1})
    assert client.safe_call('safe_call', a = 1) == {u'a': 1}
    assert client.call('echo.json', a = '1.0') == {u'a': 1}
    assert client.safe_call('safe_call.json', a = '1.0') == {u'a': 1}
    assert client.logout() is True

def test_HttpClient():
    from multiprocessing import Process
    from fanery._server import start_wsgi_server
    from fanery._client import HttpClient

    User, user = store_setUp()

    server_proc = Process(target = start_wsgi_server)
    server_proc.daemon = True
    server_proc.start()

    client = HttpClient()

    try:
        assert client.login(user.login, user.password) is True
        assert client.call('echo', a = '1.0') == repr({'a': 1})
        assert client.safe_call('safe_call', a = 1) == {u'a': 1}
        assert client.call('echo.json', a = '1.0') == {u'a': 1}
        assert client.safe_call('safe_call.json', a = '1.0') == {u'a': 1}
        assert client.logout() is True
        assert client.login(user.login, user.password) is True
        assert client.logout() is True
    finally:
        server_proc.terminate()

def main():
    from random import shuffle
    tests = [(k, v) for k, v in globals().iteritems()
             if k.startswith('test_') and callable(v)]
    shuffle(tests)

    from fanery import timecall, memory_profile, line_profile
    for _, test in tests:
        #memory_profile(test)()
        #line_profile(test)()
        timecall(test)()

if __name__ == '__main__':
    main()



__all__ = ['get_random_bytes', 'gen_token', 'hmac', 'sha256', 'AES']

from base64 import b64encode, b64decode
from hashlib import md5, sha256
import hmac as _hmac
from Crypto.Cipher import AES as _AES
from Crypto.Random import get_random_bytes

def gen_token(size = 32):
    try:
        return b64encode(get_random_bytes(size))
    except AssertionError:
        from Crypto import Random
        Random.atfork()
        return gen_token(size)

def hardened_key(key, hfun = sha256):
    return b64encode(hfun(hfun(key).digest()).digest())

def hmac(secret, text):
    return b64encode(_hmac.new(hardened_key(secret), text).digest())

def EVP_ByteToKey(password, salt, key_len=32, iv_len=16, df=md5):
    """
    Derive the key and the IV from the given password and salt.

    borrowed from
    http://stackoverflow.com/questions/13907841/implement-openssl-aes-encryption-in-python
    """
    dtot = df(password + salt).digest()
    d = [dtot]
    s_len = key_len + iv_len
    while len(dtot) < s_len:
        d.append(df(d[-1] + password + salt).digest())
        dtot += d[-1]
    return dtot[:key_len], dtot[key_len:s_len]

class AES(object):
    """
    OpenSSL compatible AES implementation.

    Interoperable with Crypto-JS
    https://code.google.com/p/crypto-js/#With_OpenSSL
    """

    def __init__(self, password, df=md5, mode=_AES.MODE_CBC):
        self.__password = hardened_key(password)
        self.__mode = mode
        self.__df = df

    def encrypt(self, text):
        salt = get_random_bytes(8)
        key, iv = EVP_ByteToKey(self.__password, salt, df=self.__df)
        l_pad = 16 - (len(text) % 16)
        t_pad = text + (chr(l_pad) * l_pad)
        cipher = _AES.new(key, self.__mode, iv)
        return b64encode('Salted__' + salt + cipher.encrypt(t_pad))

    def decrypt(self, enc):
        stream = b64decode(enc)
        assert stream[:8] == 'Salted__', 'Bad format'
        salt = stream[8:16]
        key, iv = EVP_ByteToKey(self.__password, salt, df=self.__df)
        cipher = _AES.new(key, self.__mode, iv)
        t_pad = cipher.decrypt(stream[16:])
        return t_pad[:-ord(t_pad[-1])]

def main():
    from sys import argv
    aes = AES(argv[-2])
    msg = argv[-1]
    enc = aes.encrypt(msg)
    assert msg == aes.decrypt(enc), 'failed'
    print 'OK (%f kb -> %f kb)' % (len(msg)/1024.0, len(enc)/1024.0)

if __name__ == '__main__':
    main()



from setuptools import setup

# build-essential libffi-dev python-setuptools python-dev libevent-dev libgraphviz-dev
# easy_install pip
# pip install virtualenv

setup(
    name = 'Fanery',
    version = '0.1.3',
    author = 'Marco Caramma',
    author_email = 'marco@globalsoftwaresecurity.com',
    packages = [
        'fanery',
        'tests',
        ],
    url = 'http://pypi.python.org/pypi/Fanery/',
    license = 'BSD',
    description = 'Application development framework',
    long_description = open('README.txt').read(),
    install_requires = [
        'python-dateutil',
        'memory_profiler',
        'line_profiler',
        'profilehooks',
        'pprofile',
	    'linesman',
        'objgraph',
        'psutil',
        'numpy',
        'webob',
        'ujson',
        ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
